part of fields.mirrors;

class FieldMirror<T> extends Field<T> {
  
  /// The declared name of this field.
  final Symbol declaredName;
  
  /// Returns `true` if this field is declared as a getter.
  final bool isGetter;
  
  /// Returns `true` if this field is declared as an instance variable.
  final bool isVariable;
  
  /// Creates a new field for the instance getter with the given [declaredName] 
  /// in the given [owner] instance.
  /// 
  /// An optional [getter] mirror and [parent] field may also be given.  In
  /// general these need not be specified; They are exposed as optional 
  /// parameters to allow editor implementations to optimize by caching and 
  /// reusing these values for multiple fields.
  factory FieldMirror.forGetter(
    Symbol declaredName,
    InstanceMirror owner,
   {MethodMirror getter,
    Field parent}) {
    if (getter == null) {
      getter = owner.type.declarations[declaredName];
      assert(getter != null);
      assert(getter is MethodMirror);
      assert(getter.isGetter);
      assert(!getter.isStatic);
    }
    final setter = new Symbol('${MirrorSystem.getName(declaredName)}=');
    final metadata = _parseMetadata(getter);      
    if (metadata[Ignore]) {
      return null;
    }   
    return new FieldMirror<T>._(
        declaredName,
        () => owner.getField(declaredName).reflectee,
        (value) => owner.setField(declaredName, value),
        type: _typeOf(getter.returnType),
        parent: parent,        
        editorType: metadata[FieldEditorType],
        transform: metadata[TransformType],
        annotations: metadata[FieldAnnotation],
        description: metadata[Description],
        displayName: metadata[DisplayName],
        isPrivate: getter.isPrivate,
        isReadOnly: !owner.type.declarations.containsKey(setter) 
          || metadata[ReadOnly],
        isGetter: true);
  }
  
  /// Creates a new field for the instance variable with the given 
  /// [declaredName] in the given [owner] instance.
  /// 
  /// An optional [variable] mirror and [parent] field may also be given.  In
  /// general these need not be specified; They are exposed as optional 
  /// parameters to allow editor implementations to optimize by caching and 
  /// reusing these values for multiple fields.
  factory FieldMirror.forVariable(
      Symbol declaredName,
      InstanceMirror owner,
     {VariableMirror variable,
      Field parent}) {
    if (variable == null) {
      variable = owner.type.declarations[declaredName];
      assert(variable != null);
      assert(variable is VariableMirror);
      assert(!variable.isStatic);
    }
    final metadata = _parseMetadata(variable);      
    if (metadata[Ignore]) {
      return null;
    }
    return new FieldMirror<T>._(
        declaredName,
        () => owner.getField(declaredName).reflectee,
        (value) => owner.setField(declaredName, value),
        type: _typeOf(variable.type),
        parent: parent,    
        editorType: metadata[FieldEditorType],
        transform: metadata[TransformType],
        annotations: metadata[FieldAnnotation],
        description: metadata[Description],
        displayName: metadata[DisplayName],
        isPrivate: variable.isPrivate,
        isReadOnly: variable.isFinal || metadata[ReadOnly],
        isVariable: true);
  }
      
  FieldMirror._(
      Symbol declaredName,
      T getter(),
      setter(T val),
     {Type type,
      Field parent,      
      Type editorType, 
      Transform transform,
      List annotations,
      String description,
      String displayName,
      bool isPrivate: false, 
      bool isReadOnly: false,
      this.isGetter: false,
      this.isVariable: false}) : super(
          MirrorSystem.getName(declaredName),
          getter,
          setter: setter, 
          type: type, 
          parent: parent,
          editorType: editorType,
          transform: transform,
          annotations: annotations,
          description: description,
          displayName: displayName,
          isPrivate: isPrivate,
          isReadOnly: isReadOnly)
      , this.declaredName = declaredName {
    assert(_isAssignableFrom(reflectType(T), reflectType(this.type)));
  }
}
