part of fields.model;

class Field<T> implements Function {
    
  Function _getter;
  Function _setter;
  
  final List<FieldAnnotation> annotations;
  
  FieldSet _children;
  /// The children of this field.
  FieldSet get children => _children;
  
  /// A description of this field; this is `null` by default.
  final String description;
  
  /// The display name of this field; this is equal to the `name` of this field
  /// by default.
  final String displayName;
  
  /// The type of this field's editor; this is `null` by default.
  final Type editorType;
  
  /// Returns `true` if this field is private.
  final bool isPrivate;
  
  /// Returns `true` if this field is read-only.
  final bool isReadOnly;
  
  /// The name of this field.
  final String name;

  /// The parent of this field or `null` if this field has no parent.
  final Field parent;
      
  /// This field's preferred transform; this is `null` by default.
  final Transform transform;
  
  /// The type of this field's [value].
  final Type type;
  
  T _value;
  /// The current value of this field.
  T 
    get value => _value;
    set value(T value) {
      assert(!isReadOnly);
      if (value != _value) {
        _value = value;
        _setter(value);        
        _refreshChildren().then((_) {
          _onValueChange.add(value);
          Field p = parent;
          while (p != null) {
            p._onValueChange.add(p._value);
            p = p.parent;
          }
        });
      }
    }
    
  final _onValueChange = new StreamController<T>.broadcast();
  /// An observable stream of this field's values as they change.
  Stream<T> get onValueChange => _onValueChange.stream;
      
  Field(
      String name,
      T getter(),
     {setter(T val),
      Type type,
      this.parent,
      this.editorType,
      this.transform,
      List<FieldAnnotation> annotations,
      this.description,
      String displayName,
      this.isPrivate: false,
      bool isReadOnly})      
      : this.name = name
      , this.displayName = (displayName == null) ? name : displayName
      , this._getter = getter
      , this._setter = setter
      , this.annotations = (annotations == null) ? <FieldAnnotation>[] 
          : annotations
      , this.type = (type == null) ? T : type
      , this.isReadOnly = (isReadOnly == null) ? (setter == null) : isReadOnly {
    assert(name != null);
    assert(getter != null);
    _children = new FieldSet(this);
    _value = _getter();
    _children.refreshSource(_value);
  }
  
  /// Returns this field's [value], optionally setting it first to the given 
  /// value [v] if [v] is not `null`.
  T call([T v]) {
    if (v != null) {
      value = v;      
    }
    return value;
  }
  
  /// Refresh this field's [value].
  /// 
  /// A field can only know about changes to its [value] that happen through
  /// this interface.  If another object changes the underlying value in some 
  /// other way then this method should be called.
  void refreshValue() {
    final value = _getter();
    if (value != _value) {
      _value = value;
      _refreshChildren().then((_) => _onValueChange.add(value));
    }
  }
        
  Future _refreshChildren() {
    children.refreshSource(_value);
    return Future.wait(children.map((f) => new Future(f.refreshValue)));
  }
}
