part of fields.model;

typedef Set<Field> CreateFieldSet(Object o, [Field parent]);

final Map<Type, CreateFieldSet> fieldSetFactories = {};

class FieldSet extends SetBase<Field> {
  
  static final CreateFieldSet defaultFactory = (Object o, [Field parent]) {
    if (fieldSetFactories.containsKey(o.runtimeType)) {
      return fieldSetFactories[o.runtimeType](o, parent);
    }
    return new Set<Field>();
  };
    
  static CreateFieldSet factory = defaultFactory;
    
  static final Comparator defaultComparator = 
      (Field a, Field b) => a.displayName.compareTo(b.displayName);
    
  static final Function defaultPredicate = (Field f) => true;
  
  // Consider fields in a set equal if their names and type match.
  static bool _fieldSetEquals(Field a, Field b) => 
      a.name == b.name && a.type == b.type;
  static int _fieldSetHashCode(Field a) => a.name.hashCode ^ a.type.hashCode;
  
  final Set<Field> _source = new LinkedHashSet<Field>(
      equals: _fieldSetEquals, hashCode: _fieldSetHashCode);
  final Set<Field> _view = new LinkedHashSet<Field>(
      equals: _fieldSetEquals, hashCode: _fieldSetHashCode);
  
  final List<StreamSubscription> _listeners = new List<StreamSubscription>();
  final Field _parent;
  
  final _onFieldChange = new StreamController<Field>();
  Stream<Field> get onFieldChange => _onFieldChange.stream;
  
  Comparator<Field> _comparator = defaultComparator;
  Comparator<Field> 
    get comparator => _comparator;
    set comparator(Comparator<Field> c) {
      assert(c != null);
      if (_comparator != c) {
        _comparator = c;
        for (Field f in _source) {
          f.children.comparator = c;
        }
        refreshView();
      }
    }
    
  Function _predicate = defaultPredicate;
  Function
    get predicate => _predicate;
    set predicate(bool p(Field)) {
      assert(p != null);
      if (_predicate != p) {
        _predicate = p;
        for (Field f in _source) {
          f.children.predicate = p;
        }
        refreshView();
      }
    }
  
  FieldSet([this._parent]);
  
  bool add(Field field) {
    if (!_source.add(field)) {
      return false;
    }
    refreshView();
    return true;
  }
  
  bool remove(Field field) {
    if (!_source.remove(field)) {
      return false;
    }
    refreshView();
    return true;
  }
  
  bool contains(Field field) => _view.contains(field);
  
  Field lookup(Field field) => _view.lookup(field);
  
  Iterator<Field> get iterator => _view.iterator;
  
  int get length => _view.length;
  
  Set<Field> toSet() => _view.toSet();
  
  void refreshSource(Object object) {
    final fields = factory(object, _parent);
    // Remove any fields from the source list that are not in the new object.
    _source.removeWhere((f) => 
        // Consider fields equal if their name & type match.
        !fields.any((f2) => f2.name == f.name && f2.type == f.type));
    fields.forEach((field) {
      if (_source.contains(field)) {
        _source.lookup(field)
          ..annotations.clear()
          ..annotations.addAll(field.annotations)
          .._getter = field._getter
          .._setter = field._setter
          ..refreshValue();
      }
      else {
        // Fields need to watch all of their siblings.
        fields
          .where((f) => f != field)
          .forEach((f) => f.onValueChange.listen((_) => field.refreshValue()));        
        _source.add(field);
      }
    });
    refreshView();
  }
  
  void refreshView() {    
    _listeners.forEach((s) => s.cancel());
    _listeners.clear();    
    _view
    ..clear()
    ..addAll(_source.where(predicate).toList()..sort(comparator))
    ..forEach((Field f) {
      _listeners.add(f.onValueChange.listen((_) {
        if (_onFieldChange.hasListener) {
          _onFieldChange.add(f);
        }
      }));
      f.children.refreshView();
    });
  }
}
