part of fields.model;

/// An annotation that may be applied to an instance variable or getter of an 
/// object's class declaration.
/// 
/// A default set of field annotations are defined in this library.  This
/// annotation may be extended or implemented to provide metadata to custom
/// field editors.  Custom field annotations are made available to editors as
/// [Field.annotations].
abstract class FieldAnnotation {
  const FieldAnnotation();  
}

/// An annotation that declares a field should be ignored by editors.
///
/// A call to [new Field.forVariable] or [new Field.forGetter] will return 
/// `null` if this annotation is found in the field's declaration.
const FieldAnnotation Ignore = const _Ignore();
class _Ignore extends FieldAnnotation {
  const _Ignore();
}

/// An annotation that declares a description for a field.
class Description extends FieldAnnotation {
  
  /// The description for the annotated field.
  final String description;
  
  /// Create a constant description annotation with the given [description].
  const Description(this.description);
}

/// An annotation that declares a display name for a field.
///
/// By default the display name of a field is the name of the field.
class DisplayName extends FieldAnnotation {
  
  /// The display name for the annotated field.
  final String name;
  
  /// Create a constant display name annotation with the given [name].
  const DisplayName(this.name);
}

/// An annotation that declares a field to be read-only.
/// 
/// The default value of [Field.isReadOnly] is `true` for final variables or a
/// getter with no matching setter.  This annotation provides an override to 
/// declare that a field be considered read-only if the default value is 
/// `false`.
const FieldAnnotation ReadOnly = const _ReadOnly();
class _ReadOnly extends FieldAnnotation {
  const _ReadOnly();
}

/// An annotation that declares the use of a specific [FieldEditor] for a field.
/// 
/// By default a field editor is chosen based on a field's `runtimeType`.  This
/// annotation provides an override to declare that a field use an instance of
/// the specified [editorType].
class FieldEditorType extends FieldAnnotation {
  
  /// The editor type to use for the annotated field.
  final Type editorType;
  
  /// Create a constant field editor annotation with the given [editorType].
  /// 
  /// The given [editorType] must extend or implement [FieldEditor].
  const FieldEditorType(this.editorType);
}

/// An annotation that declares the use of a specific [Transform] for a 
/// field.
/// 
/// By default a transform is chosen based on a field's `runtimeType`.  This 
/// annotation provides an override to declare that a field use the given
/// [transform] instead.
class TransformType extends FieldAnnotation {
  
  /// The transform to use for the annotated field.
  final Transform transform;  
  
  /// Create a constant transform annotation with the given [transform].
  const TransformType(this.transform);
}

class Range extends FieldAnnotation {
  
  final num min, max, step;
  const Range({this.min, this.max, this.step});
  
  /// Applies this range to the given [object].
  /// 
  /// The [object] is expected to have fields named `min`, `max`, and `step`;
  /// an example of such an object is an instance of [InputElement].
  void applyTo(object) {
    if (min != null) object.min = min.toString();
    if (max != null) object.max = max.toString();      
    if (step != null) object.step = step.toString();
  }
  
  /// Returns `true` if the given number [n] is within this range.
  bool check(num n) {
    if (n == null || n.isNaN) return false; 
    return (min == null || n >= min) && (max == null || n <= max);
  }
}

class Values extends FieldAnnotation {
  
  final List initialValues;
  final Function getValues;
  const Values(this.initialValues, {this.getValues});
  
  List get values {
    final List values = new List.from(initialValues);
    if (getValues != null) {
      values.addAll(getValues());
    }
    return values;
  }
}

/// An annotation that declares the use of another [annotation] for the items
/// of a [List] field.
class Items extends FieldAnnotation {
  
  /// The annotation to use for the items of the annotated list.
  final FieldAnnotation annotation;
  
  const Items(this.annotation);
}
