part of fields.view;

class ObjectEditorResources extends Resources {
  String get style =>
"""
:host {
  background-color: ${theme.colors.background};
  border:1px solid ${theme.colors.border};
  font-family: ${theme.display.fontFamily};
  font-size: ${theme.display.fontSize};
  height:calc(100% - 2px);      
  width:calc(100% - 2px);
}
[hidden] { display: none !important; }
button, input {
  background-color:transparent;
  border-style:none;
  color:${theme.colors.text};
}
input[type="text"] {  
  width:100%;
}
button:disabled, input[type="text"][disabled] { 
  color:${theme.colors.textDisabled};
}
input[class=error] {
  border: 2px solid ${theme.colors.error};
  margin: 0;
}
select {
  background-color:transparent;
  border: 0;
  border-radius: 0;
  color:${theme.colors.text};
}
optgroup, option {
  background: ${theme.colors.backgroundLight};
}
table {
  border-collapse:collapse;
  border-top:1px solid ${theme.colors.border};
  margin:0px;            
  padding:0px;
  table-layout:fixed;
  width:100%;
}
td {
  background-color: ${theme.colors.backgroundLight};
  border-top:1px solid ${theme.colors.border};
  color: ${theme.colors.text};
  margin:0px;
  overflow:hidden;
  padding:0px;
  text-overflow:ellipsis;
}
.expander {
  background-color:transparent;
  border-style:none;
  color:${theme.colors.text};
  flex-shrink:0;
  height:${theme.display.boxSize};   
  width:${theme.display.boxSize};
  padding:0;
}
.expander:focus {
  outline:none;
}
.editorbox {
  display:flex; 
  align-items:center;
  min-height:${theme.display.boxSize};
}
.controls {  
  display:flex; 
  height:${theme.display.boxSize};
  width:100%;
}
.filter {
  background-color:transparent;
  border-style:none;
  color:${theme.colors.text};  
}
.namebox {
  border-right:1px solid ${theme.colors.border};
  padding-left:4px;
  width:35%;  
}
.fields {
  height:calc(100% - ${theme.display.boxSize});   
  overflow-y: auto;
}
.toggle {
  background-color:transparent;
  border-style:none;
  color:${theme.colors.text};
  margin: 0px 1px 0px 0px;
}
.toggle:disabled {
  opacity:0.4;
}
.toggle[down] {
  background-color: ${theme.colors.backgroundLight};
}
""";
  const ObjectEditorResources();
}

class ObjectEditor extends EditorView<Object> with Themeable {
  
  static const TAG = 'object-editor';
  static ObjectEditorResources _factoryResources;

  String _filterText = '';  
  Function _filter;
  Function _filterPrivate;
  Function _reverseCompare;
    
  final ObjectEditorResources resources;
  final StyleElement styleElement = new StyleElement();
    
  ToggleButton _sortButton;
  ToggleButton _privateButton;
  InputElement _filterBox;
  
  FieldSet _fieldSet;
  FieldSetView _fieldSetView;
  StreamSubscription _listListener;
  
  final _onFieldChange = new StreamController<Field>.broadcast();
  Stream<Field> get onFieldChange => _onFieldChange.stream;
  
  /// Creates a new object editor with the given initial [selection] and 
  /// [resources].
  factory ObjectEditor({Object selection,
      ObjectEditorResources resources: const ObjectEditorResources()}) {    
    _factoryResources = resources;
    final editor =  new Element.tag('div', TAG);
    editor.initialize(selection);
    return editor;
  }
  
  ObjectEditor.created() 
  : super.created()
  , resources = _factoryResources;
  
  void initialize_(Object selection, _) {
    
    _filter = (Field f) => f.name.contains(_filterText);
    _filterPrivate = (Field f) => !f.isPrivate && _filter(f);
    _reverseCompare = (Field a, Field b) => a.name.compareTo(b.name) * -1;
    
    children(this, () {
      shadow(() {
        append(styleElement)..text = resources.style;
        span(() {
          _sortButton = toggleButton(() {
            span()..innerHtml = '&#9650;'..classes = ['up'];
            span()..innerHtml = '&#9660;'..classes = ['down'];
          })
          ..classes = ['toggle']
          ..disabled = true
          ..onToggle.listen((bool isDown) {
            if (isDown) {
              _fieldSet.comparator = _reverseCompare;
            } else {
              _fieldSet.comparator = FieldSet.defaultComparator;
            }
            _refresh();
          });
          _privateButton = toggleButton(() {
            span()..text = '_'..classes = ['up down'];
          })
          ..classes = ['toggle']
          ..disabled = true
          ..onToggle.listen((bool isDown) {
            if (isDown) {
              _fieldSet.predicate = _filter;
            } else {
              _fieldSet.predicate = _filterPrivate;
            }
            _refresh();
          });          
          _filterBox = input()
          ..type = 'text'
          ..classes = ['filter']
          ..disabled = true
          ..placeholder = 'filter';
        })
        ..classes = ['controls']
        ..onInput.listen((_) {
          _filterText = _filterBox.value;
          _fieldSet.refreshView();
          _refresh();
        });
        _fieldSetView = append(new FieldSetView())..classes = ['fields'];
      });
    });
  }
  
  /// Refreshes this object editor to reflect the current state of the
  /// [selection] object.
  /// 
  /// This method does _not_ need to be called for any changes made to the 
  /// [selection] object through this object editor or any field editors.  It 
  /// _does_ need to be called when the [selection] object is modified external
  /// to this library, for example by another user interface element.
  void refresh() {
    if (hasSelection) {
      _fieldSet.refreshSource(selection);
      _refresh();
    }
  }
  
  void _refresh() { 
    _sortButton.disabled = !hasSelection;
    _privateButton.disabled = !hasSelection;
    _filterBox.disabled = !hasSelection;
    _fieldSetView.refresh(_fieldSet);
  }
  
  void select_(Object value) {
    if (_listListener != null) _listListener.cancel();
    _fieldSet = new FieldSet()
    ..refreshSource(value)
    ..comparator = _sortButton.isDown ? _reverseCompare : 
        FieldSet.defaultComparator
    ..predicate = _privateButton.isDown ? _filter : _filterPrivate;      
    _listListener = _fieldSet.onFieldChange.listen((f) {
      _onFieldChange.add(f);
    });
    _fieldSetView.clear();
    _refresh();
  }
}
