@TestGroup('FieldSet')
library fields.test.model.field_set;

import 'dart:mirrors';
import 'package:bench/bench.dart';
import 'package:unittest/unittest.dart';
import 'package:fields/mirrors.dart';
import 'package:fields/model.dart';

main() => reflectTests();

// -----------------------------------------------------------------------------
// Setup
// -----------------------------------------------------------------------------

@Setup
void setup() {
  FieldSet.factory = reflectFields;
}

@Teardown
void teardown() {
  FieldSet.factory = FieldSet.defaultFactory;
}

class HasFields {
  int i = 7;
  final int ri = 42;
  num n = 3.14;
  final num rn = 6.28;
  double d = 1024.4201;
  final double rd = 1.024;
  String s = 'snarf';
  final rs = 'immutable';
  List<int> li = [ 1, 2, 3, 4];
}

class HasMetadata {
  @FieldEditorType(CustomIntEditor)    
  int customInt = 42;
  @ReadOnly
  double readOnlyDouble = 3.14;  
  @Ignore
  String secret = 'shhhh';
  @DisplayName('goodName')
  String badName = 'rename me please!';
}

class Proxy {
  final Object _o;
  final InstanceMirror _mirror;
  Proxy(Object o) : _o = o, _mirror = reflect(o);
  noSuchMethod(Invocation invocation) => _mirror.delegate(invocation);
  bool operator ==(Object other) => _o == other;
  int get hashCode => _o.hashCode;
  Type get runtimeType => _o.runtimeType;
  String toString() => _o.toString();
}

class TooManyFieldEditors {
  @FieldEditorType(CustomIntEditor)
  @FieldEditorType(CustomIntEditor)
  int customInt = 42;
}

class TooManyNames {
  @DisplayName('one')
  @DisplayName('too many')
  int customInt = 42;
}

class CustomIntEditor extends FieldEditor<int> {
  final element = null;
  CustomIntEditor(Field field) : super(field);
  void refreshElement() {}
}

// -----------------------------------------------------------------------------
// Basic
// -----------------------------------------------------------------------------

@Test('Test the construction of a field set.')
void testConstructor() {
  final set = new FieldSet();
  expect(set.comparator, equals(FieldSet.defaultComparator));
  expect(set.predicate, equals(FieldSet.defaultPredicate));
  expect(set, isEmpty);
}

@Test('Test refreshing the source of a field set with default comparator'
      ' and predicate.')
void testRefreshSourceDefaults() {
  final obj = new HasFields();
  final set = new FieldSet(); 
  set.refreshSource(obj);  
  expect(set.length, equals(9));
  expect(set.elementAt(0).name, equals('d'));
  expect(set.elementAt(1).name, equals('i'));
  expect(set.elementAt(2).name, equals('li'));
  expect(set.elementAt(3).name, equals('n'));
  expect(set.elementAt(4).name, equals('rd'));
  expect(set.elementAt(5).name, equals('ri'));
  expect(set.elementAt(6).name, equals('rn'));
  expect(set.elementAt(7).name, equals('rs'));
  expect(set.elementAt(8).name, equals('s'));
}

@Test('Test call to refreshSource with null argument.')
void testRefreshSourceNull() {
  final list = new FieldSet(); 
  list.refreshSource(null);  
  // There are not really any assumptions we can make about the fields of Null:
  // https://code.google.com/p/dart/issues/detail?id=7502
  // So we are basically just verifying that it does not throw.
}

@Test('Test that setting a null predicate on a field set throws.')
void testSetNullPredicateThrows() {
  final set = new FieldSet(); 
  expect(() {
    set.predicate = null;
  }, throwsA(const isInstanceOf<AssertionError>()));
}

@Test('Test that setting a null comparator on a field set throws.')
void testSetNullComparatorThrows() {
  final set = new FieldSet(); 
  expect(() {
    set.comparator = null;
  }, throwsA(const isInstanceOf<AssertionError>()));
}

@Test('Test setting a reverse name comparator on a field set.')
void testSetReverseNameComparator() {
  final obj = new HasFields();
  final set = new FieldSet();
  set.refreshSource(obj);
  final Comparator reverse = (Field a, Field b) => a.name.compareTo(b.name)*-1;    
  set.comparator = reverse;
  expect(set.comparator, equals(reverse));
  expect(set.length, equals(9));
  expect(set.elementAt(0).name, equals('s'));
  expect(set.elementAt(1).name, equals('rs'));
  expect(set.elementAt(2).name, equals('rn'));
  expect(set.elementAt(3).name, equals('ri'));
  expect(set.elementAt(4).name, equals('rd'));
  expect(set.elementAt(5).name, equals('n'));
  expect(set.elementAt(6).name, equals('li'));
  expect(set.elementAt(7).name, equals('i'));  
  expect(set.elementAt(8).name, equals('d'));      
}

@Test('Test setting a predicate that accepts only read-only fields.')
void testSetReadOnlyPredicate() {
  final obj = new HasFields();
  final set = new FieldSet();
  set.refreshSource(obj);
  final Function readonly = (Field f) => f.isReadOnly;
  set.predicate = readonly;
  expect(set.predicate, equals(readonly));
  expect(set.length, equals(4));
  expect(set.elementAt(0).name, equals('rd'));
  expect(set.elementAt(1).name, equals('ri'));
  expect(set.elementAt(2).name, equals('rn'));
  expect(set.elementAt(3).name, equals('rs'));
}

@Test('Test the combination of setting a reverse name comparator and a'
      ' read-only predicate on a field set.')
void testSetReverseNameComparatorAndReadOnlyPredicate() {
  final obj = new HasFields();
  final set = new FieldSet();
  set.refreshSource(obj);
  final Comparator reverse = (Field a, Field b) => a.name.compareTo(b.name)*-1;
  final Function readonly = (Field f) => f.isReadOnly;
  set.comparator = reverse;
  set.predicate = readonly;
  expect(set.length, equals(4));
  expect(set.elementAt(0).name, equals('rs'));
  expect(set.elementAt(1).name, equals('rn'));
  expect(set.elementAt(2).name, equals('ri'));
  expect(set.elementAt(3).name, equals('rd'));
}

@Test('Test setting the value of a field in the current view.')
void testSetFieldValueInView() {
  final obj = new HasFields();
  final set = new FieldSet();
  set.refreshSource(obj);  
  Field i = set.elementAt(1);
  set.onFieldChange.listen(expectAsync((f) {
    expect(f, equals(i));
  }));
  i.value = 42;
  expect(i.value, equals(42));
}

@Test('Test setting the value of a field not in the current view.')
void testSetFieldValueNotInView() {
  final obj = new HasFields();
  final set = new FieldSet();
  set.refreshSource(obj);
  Field i = set.elementAt(1);
  final Function readonly = (Field f) => f.isReadOnly;
  set.predicate = readonly;
  expect(set, isNot(contains(i)));
  set.onFieldChange.listen((_) => fail('No field in view has changed.'));
  i.value = 42;
  expect(i.value, equals(42));
}

@Test('Test refreshing the source of a field set on a dirty object.')
void testRefreshSourceDirtyObject() {
  final obj = new HasFields();
  final set = new FieldSet(); 
  set.refreshSource(obj);
  expect(set.length, equals(9));  
  set.elementAt(0).onValueChange.listen(expectAsync((value) {
    expect(value, equals(3.14));
  }));  
  // Modify a field of the object.
  obj.d = 3.14;  
  // Call to `refreshSource` on dirty object should trigger field value events.
  set.refreshSource(obj);
}

// -----------------------------------------------------------------------------
// Metadata
// -----------------------------------------------------------------------------

@Test('Test that successful parse of object field metadata.')
void testParseMetadata() {
  final obj = new HasMetadata();
  final set = new FieldSet(); 
  set.refreshSource(obj);  
  expect(set.length, equals(3));    
  expect(set.elementAt(0).declaredName, equals(#customInt));
  expect(set.elementAt(1).declaredName, equals(#badName));
  expect(set.elementAt(1).name, equals('badName'));
  expect(set.elementAt(1).displayName, equals('goodName')); 
  expect(set.elementAt(2).declaredName, equals(#readOnlyDouble));
  expect(set.elementAt(2).isReadOnly, isTrue);  
}

@Test('Test that parsing multiple @UseFieldEditor annotations throws.')
void testParseMultipleUseFieldEditorAnnotationsThrows() {
  final obj = new TooManyFieldEditors();
  final set = new FieldSet(); 
  expect(() {
    set.refreshSource(obj);
  }, throwsA(const isInstanceOf<AssertionError>()));  
}

@Test('Test that parsing multiple @Name annotations throws.')
void testParseMultipleNameAnnotationsThrows() {
  final obj = new TooManyNames();
  final set = new FieldSet(); 
  expect(() {
    set.refreshSource(obj);
  }, throwsA(const isInstanceOf<AssertionError>()));  
}

// -----------------------------------------------------------------------------
// Proxy
// -----------------------------------------------------------------------------

@Test('Test refreshing the source of a field set with a proxy object.')
void testRefreshSourceProxy() {
  final obj = new HasFields();
  final proxy = new Proxy(obj);
  final set = new FieldSet(); 
  set.refreshSource(proxy);
  expect(set.length, equals(9));  
  expect(set.elementAt(0).name, equals('d'));
  expect(set.elementAt(1).name, equals('i'));
  expect(set.elementAt(2).name, equals('li'));
  expect(set.elementAt(3).name, equals('n'));
  expect(set.elementAt(4).name, equals('rd'));
  expect(set.elementAt(5).name, equals('ri'));
  expect(set.elementAt(6).name, equals('rn'));
  expect(set.elementAt(7).name, equals('rs'));
  expect(set.elementAt(8).name, equals('s'));
}

@Test('Test setting the value of a field via a proxy object.')
void testSetFieldValueProxy() {
  final obj = new HasFields();
  final proxy = new Proxy(obj);
  final set = new FieldSet();
  set.refreshSource(proxy);  
  Field i = set.elementAt(1);
  set.onFieldChange.listen(expectAsync((f) {
    expect(f, equals(i));
  }));
  i.value = 42;
  expect(i.value, equals(42));
}
